##############################################################################
# Socks5 Client Library v1.0
#     (C)1999 Kerem 'Waster_' HADIMLI
#
# How to use:
#   1) Create your socket connected to the Socks server.
#   2) Call socks:init procedure with these 6 parameters:
#        1- Socket ID : The socket identifier that's connected to the socks5 server.
#        2- Server hostname : The main (not socks) server you want to connect
#        3- Server port : The port you want to connect on the main server
#        4- Authentication : If you want username/password authenticaton enabled, set this to 1, otherwise 0.
#        5- Username : Username to use on Socks Server if authenticaton is enabled. NULL if authentication is not enabled.
#        6- Password : Password to use on Socks Server if authenticaton is enabled. NULL if authentication is not enabled.
#   3) It'll return you a string starting with:
#        a- "OK" if successful, now you can send/receive any data from the socket.
#        b- "ERROR:$explanation" if unsuccesfull, $explanation is the explanation like "Host not found". The socket will be automatically closed on an error.
#
#
# Notes:
#   1) NEVER use file IDs instead of socket IDs!
#   2) NEVER bind the socket (fileevent) before calling socks:init procedure.
##############################################################################
#
# Author contact information:
#   E-mail :  waster@iname.com
#   ICQ#   :  27216346
#
##############################################################################
#

proc socks:init {sck addr port auth user pass} {
set ver "\x05"               ;#Socks version
if {$auth==0} {set method "\x00"; set nmethods "\x01"} \
  elseif {$auth==1} {set method "\x00\x02"; set nmethods "\x02"} \
  else {return "ERROR:"}
set nomatchingmethod "\xFF"  ;#No matching methods

set cmd_connect "\x01"  ;#Connect command
set rsv "\x00"          ;#Reserved
set atyp "\x03"         ;#Address Type (domain)
set dlen "[binary format c [string length $addr]]" ;#Domain length (binary 1 byte)
set port [binary format S $port] ;#Network byte-ordered port (2 binary-bytes)

set authver "\x01"  ;#User/Pass auth. version
set ulen "[binary format c [string length $user]]"  ;#Username length (binary 1 byte)
set plen "[binary format c [string length $pass]]"  ;#Password length (binary 1 byte)

set a ""

#
# Because SOCKS protocol is binary (you sometimes send "\x00"), we configure our socket
#
  fconfigure $sck -translation {binary binary}

  puts -nonewline $sck "$ver$nmethods$method"
  flush $sck

  while {1} {
	catch {set a [read $sck]}
	if {$a!=""} {break}
	if {[eof $sck]} {catch {close $sck}; return "ERROR:Connection closed with Socks Server!"}
  }

  set serv_ver ""; set method $nomatchingmethod
  binary scan $a "cc" serv_ver smethod

  if {$serv_ver!=$ver} {catch {close $sck}; return "ERROR:Socks Server isn't version 5!"}

  if {$smethod=="\x00"} {} \
  elseif {$smethod=="\x02"} {  ;#User/Pass authorization required
	if {$auth==0} {catch {close $sck}; return "ERROR:Method not supported by Socks Server!"}

	puts -nonewline $sck "$authver$ulen$user$plen$pass"
	flush $sck

	while {1} {
		catch {set a [read $sck]}
		if {$a!=""} {break}
		if {[eof $sck]} {catch {close $sck}; return "ERROR:Connection closed with Socks Server!"}
	}

	set auth_ver ""; set status "\x00"
	binary scan $a "cc" auth_ver status

	if {$auth_ver!=$authver} {catch {close $sck}; return "ERROR:Socks Server's authenciation isn't supported!"}
	if {$status!="\x00"} {catch {close $sck}; return "ERROR:Wrong username or password!"}

  } else {catch {close $sck}; return "ERROR:Method not supported by Socks Server!"}

#
# We send request4connect
#
  puts -nonewline $sck "$ver$cmd_connect$rsv$atyp$dlen$host$port"
  flush $sck

  while {1} {
	catch {set a [read $sck]}
	if {$a!=""} {break}
	if {[eof $sck]} {catch {close $sck}; return "ERROR:Connection closed with Socks Server!"}
  }

  set serv_ver ""; set rep ""
  binary scan cc $a serv_ver rep

  if {$serv_ver!=$ver} {catch {close $sck}; return "ERROR:Socks Server isn't version 5!"}

  if {$rep=="\x00"} {fconfigure $sck -translation {auto auto}; return "OK"} \
    elseif {$rep=="\x01"} {catch {close $sck}; return "ERROR:SOCKS server responded:general SOCKS server failure"} \
    elseif {$rep=="\x02"} {catch {close $sck}; return "ERROR:SOCKS server responded:connection not allowed by ruleset"} \
    elseif {$rep=="\x03"} {catch {close $sck}; return "ERROR:SOCKS server responded:Network unreachable"} \
    elseif {$rep=="\x04"} {catch {close $sck}; return "ERROR:SOCKS server responded:Host unreachable"} \
    elseif {$rep=="\x05"} {catch {close $sck}; return "ERROR:SOCKS server responded:Connection refused"} \
    elseif {$rep=="\x06"} {catch {close $sck}; return "ERROR:SOCKS server responded:TTL expired"} \
    elseif {$rep=="\x07"} {catch {close $sck}; return "ERROR:SOCKS server responded:Command not supported"} \
    elseif {$rep=="\x08"} {catch {close $sck}; return "ERROR:SOCKS server responded:Address type not supported"} \
      else {catch {close $sck}; return "ERROR:SOCKS server responded: unknown error"}
}
